@@ -476,6 +476,12 @@ proc_status_t builtin_run(parser_t &parser, const wcstring_list_t &argv, io_stre
         }
         if (code < 0) {
             FLOGF(warning, "builtin %ls returned invalid exit code %d", cmdname.c_str(), code);
+            // If the code is below 0, constructing a proc_status_t
+            // would assert() out, which is a terrible failure mode
+            // So instead, what we do is we get a positive code,
+            // and we avoid 0.
+            code = abs((256 + code) % 256);
+            if (code == 0) code = 255;
         }
         return proc_status_t::from_exit_code(code);
     }
