@@ -106,6 +106,7 @@ static constexpr const input_function_metadata_t input_function_metadata[] = {
     {L"cancel", readline_cmd_t::cancel},
     {L"cancel-commandline", readline_cmd_t::cancel_commandline},
     {L"capitalize-word", readline_cmd_t::capitalize_word},
+    {L"clear-screen", readline_cmd_t::clear_screen_and_repaint},
     {L"complete", readline_cmd_t::complete},
     {L"complete-and-search", readline_cmd_t::complete_and_search},
     {L"delete-char", readline_cmd_t::delete_char},
@@ -128,6 +129,7 @@ static constexpr const input_function_metadata_t input_function_metadata[] = {
     {L"forward-single-char", readline_cmd_t::forward_single_char},
     {L"forward-word", readline_cmd_t::forward_word},
     {L"history-pager", readline_cmd_t::history_pager},
+    {L"history-pager-delete", readline_cmd_t::history_pager_delete},
     {L"history-prefix-search-backward", readline_cmd_t::history_prefix_search_backward},
     {L"history-prefix-search-forward", readline_cmd_t::history_prefix_search_forward},
     {L"history-search-backward", readline_cmd_t::history_search_backward},
@@ -464,19 +466,24 @@ class event_queue_peeker_t {
 
     /// Check if the next event is the given character. This advances the index on success only.
     /// If \p timed is set, then return false if this (or any other) character had a timeout.
-    bool next_is_char(wchar_t c, bool timed = false) {
+    bool next_is_char(wchar_t c, bool escaped = false) {
         assert(idx_ <= peeked_.size() && "Index must not be larger than dequeued event count");
         // See if we had a timeout already.
-        if (timed && had_timeout_) {
+        if (escaped && had_timeout_) {
             return false;
         }
         // Grab a new event if we have exhausted what we have already peeked.
         // Use either readch or readch_timed, per our param.
         if (idx_ == peeked_.size()) {
             char_event_t newevt{L'\0'};
-            if (!timed) {
-                newevt = event_queue_.readch();
-            } else if (auto mevt = event_queue_.readch_timed()) {
+            if (!escaped) {
+                if (auto mevt = event_queue_.readch_timed_sequence_key()) {
+                    newevt = mevt.acquire();
+                } else {
+                    had_timeout_ = true;
+                    return false;
+                }
+            } else if (auto mevt = event_queue_.readch_timed_esc()) {
                 newevt = mevt.acquire();
             } else {
                 had_timeout_ = true;
@@ -541,7 +548,7 @@ static bool have_mouse_tracking_csi(event_queue_peeker_t *peeker) {
     // Maximum length of any CSI is NPAR (which is nominally 16), although this does not account for
     // user input intermixed with pseudo input generated by the tty emulator.
     // Check for the CSI first.
-    if (!peeker->next_is_char(L'\x1b') || !peeker->next_is_char(L'[', true /* timed */)) {
+    if (!peeker->next_is_char(L'\x1b') || !peeker->next_is_char(L'[', true /* escaped */)) {
         return false;
     }
 
@@ -592,8 +599,8 @@ static bool try_peek_sequence(event_queue_peeker_t *peeker, const wcstring &str)
     for (wchar_t c : str) {
         // If we just read an escape, we need to add a timeout for the next char,
         // to distinguish between the actual escape key and an "alt"-modifier.
-        bool timed = prev == L'\x1B';
-        if (!peeker->next_is_char(c, timed)) {
+        bool escaped = prev == L'\x1B';
+        if (!peeker->next_is_char(c, escaped)) {
             return false;
         }
         prev = c;
